/**
 * Copyright (c) 2006-2009 Institut National de Recherche en Informatique et 
 *                         en Automatique. All rights reserved.
 * $COPYRIGHT$
 * 
 * Additional copyrights may follow
 * 
 * $HEADER$
 */

#define _BSD_SOURCE 1
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <unistd.h>
#include <sys/param.h>
#include <stdio.h>
#include <stdarg.h>
#include <time.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <signal.h>
#include <ctype.h>
#include <netdb.h>
#include <pthread.h>
#include <pcap.h>
#include <net/ethernet.h>
#include <sys/socket.h>
#include <netinet/in_systm.h>
#include <net/if.h>
#include <netinet/tcp.h>
#include <netinet/ip.h>
#include "../proxy/util.h"
#include "connection-helper.h"
#include "cmdline.h"

#include "stubs/serviceH.h"
#include "stubs/service.nsmap"



static struct gengetopt_args_info ARGS;

static FILE* log_file = NULL;
static int running = 0;
struct Namespace *namespaces;


char host_name[HOST_NAME_MAX];
char host_ip[MAX_IP];
char node_id[MAX_ID];
int node_port;



struct soap helper_soap;
char frontale_addr[MAX_SERVICE_ADDR];

static int log_init(const char *file);
static void log_close();
static void log_add(LOG_LEVEL level, char *fmt, ...);
static void setup_signals(void);
static void interrupt_catch(int signo);
static void startDaemon();
static void callback(u_char *, const struct pcap_pkthdr *, const u_char *);
static int sniff_connect(struct sockaddr_in local_addr, int port, struct connect_struct * connect);

int main(int argc, char *argv[])
{
  struct hostent *host = NULL;
  pid_t pid, sid; 

  if( cmdline_parser (argc, argv, &ARGS) != 0)
    exit(1) ;

if( ! ARGS.nodaemon_given ) {
  pid = fork(); 

  if (pid < 0)  exit(1);

  if (pid > 0)  exit(EXIT_SUCCESS);

  umask(0);
  sid = setsid();

  if (sid < 0)  exit(1);

  close(STDIN_FILENO);
  close(STDOUT_FILENO);
  close(STDERR_FILENO);    

  {
    FILE *f = fopen("/var/run/connection-helper.pid", "w");
     if(f != NULL)
      {
        fprintf(f, "%d\n", getpid());
        fclose(f);
	} else {
	    f = fopen("/tmp/connection_helper.pid", "w");
	    if(f != NULL) {
		fprintf(f, "%d\n", getpid());
		fclose(f);
	    }
      }
  }
 
  log_init(LOG_FILE);
}
  setup_signals();
  sprintf(frontale_addr,"http://%s:%d/",ARGS.frontal_host_arg, ARGS.frontal_port_arg);   
  if(gethostname(host_name, HOST_NAME_MAX) == -1)
    {
      perror("unable to resolve my hostname");
      return ERROR;
    }
   
  host = gethostbyname(host_name);
  if( host == NULL)
  {
      perror("gethostbyname failed");
      return ERROR;
  }
     
  strncpy(host_ip, inet_ntoa(*( struct in_addr*)( host->h_addr_list[0])), MAX_IP);
  strncpy(host_name, host->h_name, MAX_HOST_NAME);
  
  log_add(logMessage,"Daemon avant start\nHost_name = %s Host_ip = %s", host_name, host_ip);
  startDaemon();
  log_close();
  return 1;
}

static void  * handleRequest(void *p_tsoap)
{
  struct soap *v_soap=(struct soap *)p_tsoap;

  pthread_detach(pthread_self());
  service_serve((struct soap*)v_soap);
  soap_destroy((struct soap*)v_soap);
  soap_end((struct soap*)v_soap);
  soap_done((struct soap*)v_soap);
  free(v_soap);
  return NULL;

}

void startDaemon()
{
  struct soap *tmpsoap;
  pthread_t tid;

  running = 1;
  
  soap_init(&helper_soap);
  soap_set_namespaces(&helper_soap, service_namespaces);

  if(soap_bind(&helper_soap, NULL, ARGS.service_port_arg, 100) < 0)
    return;
    
  while(running)
  {
    log_add(logMessage, "Wating for clients requests ..... ");
    helper_soap.accept_timeout = 60;  // non blocking
    if(soap_accept(&helper_soap) < 0) 
      continue;
    tmpsoap = soap_copy(&helper_soap);
    pthread_create(&tid, NULL, handleRequest, (void*)tmpsoap);

  }
  soap_done(&helper_soap);
  return;


}

int log_init(const char *file)
{
   char file_name[MAX_FILE_NAME_SIZE];
   sprintf(file_name,"%s.log",file);
   log_file = fopen(file_name, "a+");
   if (!log_file)
   {
      log_add(logError, "log_init - fopen failed to open %s (%s)", file_name, strerror(errno));
      return ERROR;
   }
   return SUCCESS;
}

void log_close()
{
     if(log_file)
     {
               fflush(log_file);
               fclose(log_file);
               log_file=NULL;
     }
}

static char* get_header(LOG_LEVEL level)
{
     switch(level)
     {
          default:
          case logMessage:
          return "Message";
          case logError:
          return "Error";
     }
}

void log_add(LOG_LEVEL level, char *fmt, ...)
{
     char buffer[MAX_MSG_SIZE];
     va_list args;
     time_t t = time(NULL);
     char timeStr[256];
     
     va_start( args, fmt );
     vsprintf( buffer, fmt, args );
     va_end( args );

     strncpy(timeStr, ctime(&t), 255);
     timeStr[strlen(timeStr) - 1] = '\0';

     if (log_file)
      fseek(log_file, 0L, SEEK_END);
     fprintf(stdout, "[%s : %s] %s \n", get_header(level), timeStr, buffer);
     if(level == logError)
           fprintf(stderr, "[%s : %s] %s \n", get_header(level), timeStr, buffer);
     else
           fprintf((log_file ? log_file : stderr), "[%s : %s] %s \n", get_header(level), timeStr, buffer);
      
     if (log_file)
       fflush(log_file);
}
int ns__service_proxy_port(struct soap *p_soap, int *service_proxy_port, struct ns__noResponse  *out)
{
  return SOAP_OK;
}

void setup_signals(void)
{
     struct sigaction    sa;
     
     sa.sa_handler = SIG_IGN;                /* SIGPIPE to be ignored         */
     sigemptyset(&sa.sa_mask);               /* No signals to mask off        */
     sa.sa_flags = 0;                        /* No flags needed               */
     
     if (sigaction(SIGPIPE, &sa, NULL) == -1)
          exit(1);
     
     sa.sa_handler = interrupt_catch;        /* SIGINT signal handler         */
     sigemptyset(&sa.sa_mask);               /* No signals to mask off        */
     sa.sa_flags = SA_RESTART;               /* Restart interrupted syscalls  */
     
     if (sigaction(SIGINT, &sa, NULL) == -1)
          exit(1);
} 

void interrupt_catch(int signo)
{
     running = 0;
}



int ns__update_frontale(struct soap *soap1, int *number, struct ns__noResponse *out)
{
  return SOAP_OK;
}
int ns__proxy_port(struct soap *p_soap, int *proxy_port, struct ns__noResponse  *out)
{
  return SOAP_OK;

}
int ns__getHostURI(struct soap *soap, process_name_t *proc_name, process_name_t *my_name, sockaddr_t *uri_A, TAG_t tag, ns__uriResponse *uri_B)
{
  struct soap frsoap;
  ns__uriResponse uri_rep;

  soap_init(&frsoap);
  soap_set_namespaces( &frsoap, service_namespaces);
  soap_call_ns__getHostURI(&frsoap, frontale_addr, "",proc_name, my_name, uri_A, tag, &uri_rep);
  if(frsoap.error)
    {
      soap_print_fault(&frsoap, stderr);
      return SOAP_ERR;;
    }
  log_add(logMessage, "ns__getHostURI {%s} {%s}:{%d}  traversing_code = {%d} reponse = {%s}\n", uri_rep.host_name, uri_rep.host_ip, uri_rep.port, uri_rep.traversing_code, uri_rep.response);

  memset(uri_B , 0, sizeof(struct sockaddr_t));
  uri_B -> host_name = (char *)malloc(MAX_HOST_NAME * sizeof(char));
  memset(uri_B -> host_name, 0, MAX_HOST_NAME);

  uri_B -> host_ip = (char *)malloc(MAX_IP * sizeof(char));
  memset(uri_B ->host_ip, 0, MAX_IP);

  strncpy(uri_B ->host_ip, uri_rep.host_ip, MAX_IP);
  strncpy(uri_B ->host_name, uri_rep.host_name, MAX_HOST_NAME);

  if( uri_rep.response != NULL)
    {
      uri_B -> response = (char *)malloc(MAX_RESPONSE * sizeof(char));
      log_add(logMessage, "ns__getHostURI strncpy(uri_B -> response, %s)", uri_rep.response);
      strncpy(uri_B -> response, uri_rep.response, MAX_RESPONSE);
    }
  else
      log_add(logMessage, "ns__getHostURI response == NULL)");
  log_add(logMessage, "ns__getHostURI  uri_B -> port = %d", uri_rep.port);
  uri_B -> port = uri_rep.port;
  log_add(logMessage, "ns__getHostURI uri_B -> traversing_code = %d", uri_rep.traversing_code);
  uri_B -> traversing_code = uri_rep.traversing_code;
  if(uri_B -> traversing_code == tcp_traversing)
    {
      pid_t pid;
      struct connect_struct connect;
      struct sockaddr_in my_addr;

      my_addr.sin_family= AF_INET;
      my_addr.sin_port=htons(uri_A -> port);
      my_addr.sin_addr.s_addr =inet_addr(uri_A ->host_ip);

      switch(pid = fork())
	{
	case -1:
	  perror("fork () failed");
	  return SOAP_OK;
	case  0:
	  {
	    sniff_connect(my_addr, uri_A -> port, &connect);
	    exit(0);
	  }
	default:
	  break;

	}
    }
  log_add(logMessage, "Before return from ns__getHostURI B %s %s:%d %s %d\n", uri_B->host_name, uri_B->host_ip, uri_B->port, uri_B -> response, uri_B -> traversing_code);
  return SOAP_OK;

}
int ns__putMyURI(struct soap *soap, process_name_t *proc_name, sockaddr_t *uri_B, TAG_t tag, struct ns__noResponse  *out)
{
  struct soap frsoap;
  
  soap_init(&frsoap);
  soap_set_namespaces( &frsoap, service_namespaces);
  uri_B -> response = (char *)malloc(MAX_RESPONSE * sizeof(char));
  sprintf(uri_B -> response, "%d.%d.%d.%d %s %d", (int)proc_name->cellid,(int)proc_name->jobid,(int)proc_name->vpid, tag, host_name, ARGS.service_port_arg); 
  sprintf(node_id,"%d.%d.%d.%d",(int)proc_name->cellid,(int)proc_name->jobid,(int)proc_name->vpid, tag);
  node_port = uri_B -> port;
  log_add(logMessage,"Connection-helper putMyUri: %s:%s", uri_B -> response,frontale_addr);
  soap_call_ns__putMyURI(&frsoap, frontale_addr, "", proc_name, uri_B, tag, out);
  if(frsoap.error)
    {
      soap_print_fault(&frsoap, stderr);
      return SOAP_ERR;;
    }
  soap_end(&frsoap);
  return SOAP_OK;
}
int ns__update_broker(struct soap *soap, int *number, struct ns__noResponse  *out)
{
  return SOAP_OK;
}
int ns__update_connection_helper(struct soap *soap, int *number, struct ns__noResponse  *out)
{
  printf("ns__update_connection_helper send me the number %d\n",*number);
  return SOAP_OK;
}
int ns__proxy_connection_demande(struct soap *soap, char *demande, struct ns__noResponse  *out)
{
  char proxy_ip[MAX_IP];
  char buff[MAX_CMD_SIZE];
  int nwrite = 0;
  int proxy_port, uid;
  struct sockaddr_in serv_addr;
  int sockfd;
    
  sscanf(demande, "%d %s %d", &uid, proxy_ip, &proxy_port);
  log_add(logMessage,"connection_helper_ns__proxy_connection_demande: uid = %d" , uid);
  if((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
  {
    log_add(logMessage,"connection_helper_ns__proxy_connection_demande: socket() failed");
    return SOAP_ERR;
  }
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_addr.s_addr = inet_addr("127.0.0.1");
  serv_addr.sin_port = htons(node_port);
  log_add(logMessage,"Connecting to localhost:%d", node_port);
  if (connect(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
  {
    log_add(logError,"Connect() failed:");
    return SOAP_ERR;
  }
    
  sprintf(buff, "%d %s %d", uid, proxy_ip, proxy_port);
  log_add(logMessage,"connect() ok, sending to host %s", buff);
  nwrite = writeCmd(sockfd, buff);
  close(sockfd);
  log_add(logMessage,"write to host %d", nwrite);
  return SOAP_OK;
}

int ns__get_proxy_uid(struct soap *p_soap, int *uid)
{
  return SOAP_OK;
}
int ns__forward_syn_data(struct soap *p_soap, char *data, struct ns__noResponse  *out)
{
    return SOAP_OK;
}
int ns__start_traversing(struct soap *p_soap, char *local_ip, int *port, struct ns__noResponse  *out)
{
  return SOAP_OK;
}

int sniff_connect(struct sockaddr_in local_addr, int port, struct connect_struct * connect) 
{
  pcap_t *handle;                                    /* Session handle */
  char *dev;                                         /* The device to sniff on */
  char errbufpcap[PCAP_ERRBUF_SIZE];                 /* Error string */
  struct bpf_program filter;                         /* The compiled filter */
  char filter_app[80];
  bpf_u_int32 mask;                                  /* Our netmask */
  bpf_u_int32 net=0;                                   /* Our IP */
  pcap_if_t *alldevsp=NULL, *pdev=NULL;
  pcap_addr_t *paddr=NULL;

  if (geteuid() != 0)
    {
      log_add(logError, "pcap needs root rights");
      return ERROR;      
    }
     
  sprintf(filter_app, "tcp src port %d", port);
  if(pcap_findalldevs(&alldevsp,errbufpcap)==-1)
    {
      perror("pcap_findalldevs error");
      return ERROR;
    }

  log_add(logMessage,"SEARCH FOR %s ", inet_ntoa(local_addr.sin_addr));
  pdev=alldevsp;
  dev=NULL;
  while(!dev && pdev)
    {

      paddr=pdev->addresses;
      log_add(logMessage,"DEVICE NAME %s",pdev->name);
      while(paddr)
	{
	  //if(!memcmp(paddr->addr,&local_addr,sizeof(struct sockaddr_in)))
	  log_add(logMessage,"ADDRESS %s ", inet_ntoa(((struct sockaddr_in *)paddr->addr)->sin_addr));
	  if (((struct sockaddr_in *)paddr->addr)->sin_addr.s_addr == local_addr.sin_addr.s_addr)
	    {
	      dev=(char *)malloc(strlen(pdev->name)+1);
	      strcpy(dev,pdev->name);
	      break;
	    }
	  paddr=paddr->next;
	}
      pdev=pdev->next;

    }
  pcap_freealldevs(alldevsp);
  if(!dev)
    {

      log_add(logError,"pcap, could not find device");
      dev = pcap_lookupdev(errbufpcap);
      //log_add(logMessage,"DEV=%s",dev);
      if (pcap_lookupnet(dev, &net, &mask, errbufpcap) == -1) 
	{
          log_add(logError, "pcap_lookupnet error");
          return ERROR;
	} 
    }
  //dev="lo";

  /* Open the session in promiscuous mode */
  handle = pcap_open_live(dev, TAILLE_PAQUETS, IFF_PROMISC, 1000, errbufpcap);
     
  /* Compile and apply the filter */
  if(pcap_compile(handle, &filter, filter_app, 0x100, 0) < 0)
    {
          
      log_add(logError, "pcap_compile error: %s", pcap_geterr(handle));
      return ERROR;
    }
  if (pcap_setfilter(handle, &filter) < 0 ) 
    {
      log_add(logError, "pcap_setfilter error: %s", pcap_geterr(handle));
      return ERROR;
    } 
  /*{
    FILE *ff;
    ff=fopen("/tmp/ok","w+");
    fclose(ff);
    }*/
     
     
  if( pcap_loop(handle, 1, callback, (u_char *)connect) < 0 )
    {
      log_add(logError, "pcap_loop: %s",pcap_geterr(handle));
      return ERROR;
    }

  pcap_close(handle);
  return SUCCESS;

}


void callback(u_char *mon_param, const struct pcap_pkthdr * entete, const u_char * data)
{

  const struct ip     *paquet_ip=NULL;
  const struct tcphdr *paquet_tcp=NULL;
  
  const char * tcp_data=NULL;


  struct connect_struct *connect = (struct connect_struct *)mon_param;


  char ip_src[MAX_IP],     ip_dst[MAX_IP];
  int port_src, port_dst;

  int rst, syn, ack, fin;

  u_int ip_head_len=0, tcp_head_len=0, tcp_data_len=0, tcp_data_len_reel=0;
  u_int taille=0;
  u_int off=0;

  /*************************************************
   * PARTIE ETHERNET
   *************************************************/

  // nous sommes sur un brin ethernet.
  struct ether_header * entete_ethernet = (struct ether_header *) data;
  log_add(logMessage,"CALLBACK");

     
  // Verifier que nous avons a faire a un paquet ip.
  if(!(ntohs(entete_ethernet->ether_type) == ETHERTYPE_IP)) return;


  /*************************************************
   * PARTIE IP
   *************************************************/

  // Pointer sur la partie ip
  taille=( entete->caplen - sizeof(struct ether_header ));
  if( taille < sizeof( struct ip ))
    {
      log_add(logError,"Err Paquet tronque, partie IP incomplete");
      return;
    }

  paquet_ip=(struct ip *)(data+sizeof(struct ether_header));

  // Verifier que le paquet Ip fait bien la taille indiquee dans l'entete
  strncpy(ip_src,(char *)inet_ntoa(paquet_ip->ip_src),MAX_IP);
  strncpy(ip_dst,(char *)inet_ntoa(paquet_ip->ip_dst),MAX_IP);

  // Verifier la taille de l'entete IP.
  ip_head_len = paquet_ip->ip_hl * 4;
  if (ip_head_len < sizeof (struct ip))
    {
      log_add(logError,"Err longueur entete ip");
      return;
    }

  /* POUR DEBUG
   * fprintf(stdout,"IP_HEAD_LEN : %u - %uETH_HDR_LEN : %d",
   *ip_head_len,
   *(paquet_ip->ip_hl << 2),
   *sizeof(struct ether_header));
   */

  /*************************************************
   * PARTIE TCP
   *************************************************/

  // Verifier que la paquet IP vehicule du protocole TCP.
  off=ntohs(paquet_ip->ip_off);
  if (!(off & 0x1fff) && (paquet_ip->ip_p==IPPROTO_TCP))
    {
      // Decouper le protocole TCP.
      // fprintf(stdout,"Ethernet -> IP -> TCP ");

      // Verifier si nous avons l'enete tcp complete.
      taille=(entete->caplen - sizeof(struct ether_header ) - ip_head_len);
      if( taille < sizeof( struct tcphdr ))
	{
	  log_add(logError,"Err Paquet tronque, partie TCP incomplete");
	  return;
	}

      // Pointer sur la partie Tcp.
      paquet_tcp=(const struct tcphdr *)(data+sizeof(struct ether_header)+ip_head_len);
      port_src=ntohs(paquet_tcp->th_sport);
      port_dst=ntohs(paquet_tcp->th_dport);
      tcp_head_len=paquet_tcp->th_off*4;
      //fprintf(stdout,"Src_Port : %dDst_Port : %d",ntohs(paquet_tcp->th_sport),ntohs(paquet_tcp->th_dport));

      // Recuperer les flags Tcp.
      if (paquet_tcp->th_flags & TH_RST) rst=1; else rst=0;
      if (paquet_tcp->th_flags & TH_FIN) fin=1; else fin=0;
      if (paquet_tcp->th_flags & TH_SYN) syn=1; else syn=0;
      if (paquet_tcp->th_flags & TH_ACK) ack=1; else ack=0;

      // Verifier si il y a des donnees dans le paquet.
      tcp_data_len = entete->caplen - ( sizeof(struct ether_header) + ip_head_len + tcp_head_len );
      tcp_data_len_reel = entete->len - ( sizeof(struct ether_header) + ip_head_len + tcp_head_len );

      if(tcp_data_len>0)
	{
          log_add(logError,"TOTAL DATA : %u",tcp_data_len);
          tcp_data = ( const char * )(data+sizeof(struct ether_header)+ip_head_len+tcp_head_len);
	}

      connect->sport = ntohs(paquet_tcp->th_sport);
      connect->dport = ntohs(paquet_tcp->th_dport);
      connect->seq   = ntohl(paquet_tcp->th_seq);
      log_add(logError,"SEQUANCE = %lu",ntohl(paquet_tcp->th_seq));

      connect->saddr = paquet_ip->ip_src;
      connect->daddr = paquet_ip->ip_dst;
     
     
    }
  else
    {
      // Proto non pris en charge par mon code ...
      log_add(logError,"Ethernet -> IP -> ???? "); 
    }

  
}
int  ns__delete_data(struct soap *p_soap,int *taskid)                                                              
{                                                                                                                  
    struct soap frsoap;
   
    soap_init(&frsoap);  
    soap_set_namespaces( &frsoap, service_namespaces); 
    soap_call_ns__delete_data(&frsoap, frontale_addr, "", taskid);
    if(frsoap.error)
	{    
	    soap_print_fault(&frsoap, stderr);
	    return SOAP_ERR;;  
	}  
    return SOAP_OK;                                                                                                
}
