/**
 * Copyright (c) 2006-2009 Institut National de Recherche en Informatique et 
 *                         en Automatique. All rights reserved.
 * $COPYRIGHT$
 * 
 * Additional copyrights may follow
 * 
 * $HEADER$
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <string.h>
#include <ctype.h>
 
#include "proxy.h"
#include "util.h"

static char* nextSpace(char* buff)
{
  char *p = buff;
  while(*p && !isspace(*p))
    p++;

  return p;
}

static char* nextToken(char* buff)
{
  char *p = buff;
  while(*p && isspace(*p))
    p++;

  return p;
}

int readCmd(int sockfd, char *cmd_buff)
{
  size_t toread = MAX_CMD_SIZE;
  size_t nread_tot, nread;

  toread = MAX_CMD_SIZE;
  nread_tot = 0;
  while( (toread > 0) && (nread = recv(sockfd, cmd_buff + nread_tot, toread, 0) ) > 0)
      {
	  
	  /*int i;
	  fprintf(stderr, "----------\n");
	  for(i = 0; i < nread; i++)
	      fprintf(stderr, "%c", cmd_buff[i+nread_tot]);
	  fprintf(stderr, "\n----------\n");
	  //fprintf(stderr, "ReadCmd %d:%s",nread, cmd_buff);
	  */
      toread -= nread;
      nread_tot += nread;
    }
    
  // if( nread > 0 ) return MAX_CMD_SIZE;

  return nread_tot;
}

int writeCmd(int sockfd,char *cmd_buff)
{                         
  size_t nwrite;
  size_t ncount = 0;
  while( (ncount < MAX_CMD_SIZE) && ((nwrite = send(sockfd, cmd_buff+ncount, MAX_CMD_SIZE-ncount, 0))>0) )
    ncount+=nwrite;
  if( nwrite < 0 )
    return nwrite;
  return ncount;
}

int proxyWrite(int sockfd, char *buff, int size)
{                         
  int nwrite;
  int ncount=0;
  
  while((ncount < size) && ((nwrite = send(sockfd, buff + ncount, size - ncount, 0)) > 0 ))
    ncount+=nwrite;
  return ncount;
}

int getCmdCode(char *cmd_buff)
{
  int code;
  if(sscanf(cmd_buff,"%d",&code)<1)
    return ERROR;
  return code;
  
}
int getCmdArgc(char *cmd_buff)
{
  int count = 0;
  char *p = nextSpace(cmd_buff);

  while(*p)
  {
    p = nextToken(p);
    if (p)
      count++;
    p = nextSpace(p);
  }

  return count;
}

void getCmdArgs(char *cmd_buff, char ***args, int *argc)
{
  int i, count = 0;
  char *p = nextSpace(cmd_buff);
  *argc = getCmdArgc (cmd_buff);
  if (!(*argc))
    return;
  *args = (char **)malloc((*argc) * sizeof(char *));
   for(i = 0; i < *argc; i++)
     (*args)[i] = malloc(MAX_ARG_SIZE);

  while(*p)
  {
    p = nextToken(p);
    if (p)
    {
      char *s = nextSpace(p);
      strncpy((*args)[count], p, (s - p > MAX_ARG_SIZE - 1 ? MAX_ARG_SIZE - 1 : s - p));
      (*args)[count][s - p] = '\0';
      count++;

      p = s;
    }    
  }       
}

void freeCmdArgs(char ***args, int argc)
{
  int i;
  if (!args || !(*args))
    return;

  for(i = 0; i < argc; i++)
  {
    //fprintf(stderr,"args[%d]=%s %d \n",i,(*args)[i],argc);
    free((*args)[i]);
    (*args)[i] = NULL;
  }
  free(*args);
  *args = NULL;
}


